// LCD_PIC_I2C_GENERIC_DRIVER FOR MICROCHIP PIC
// Tested on PIC18F4550
// Some commands may be unique to CCS compiler v4.141
//
// Written by: MDP
// REV-9
// 10 May 2013
//
// WRITES to LCD I2C Adapters that use I2C chip PCF8574P
// Does not READ
//
// Put following line in main program AFTER I2C interface definition:
// #include <LCD_I2C_GENERIC_DRIVER> 
//
//
//
// Following LCD subroutines may be useful in main program:
/*
   LCD_PIC_I2C_Write_Generic(char text[]);   // writes text string 'text[]'
   LCD_PIC_I2C_Init_Generic();                       // Hitachi 44780U initialization commands 
   LCD_PIC_I2C_GOTOXY_Generic(int8 x, int8 y);       // puts cursor at position (x,y) on LCD
   LCD_PIC_I2C_Putc_Generic(char c);         // writes single character 'c'
   LCD_PIC_I2C_Send_Byte_Generic(int8 address, int8 n);  // sends arbitraty byte 'n' to address '0' - useful for special commands/characters
*/

// Following line tested with CCS compiler on a PIC18F4550 and best goes into main program body
// #use i2c(MASTER, FAST, SCL=PIN_A3, SDA=PIN_A2, FORCE_SW, STREAM=LCDI2C)
// Note: Using PICs  with HW I2C capability can increase speed up to 2x

// Line addresses for LCDs which use
// the Hitachi HD44780U controller chip
#define LCD_PIC_LINE_1_ADDRESS 0x00
#define LCD_PIC_LINE_2_ADDRESS 0x40
#define LCD_PIC_LINE_3_ADDRESS 0x14
#define LCD_PIC_LINE_4_ADDRESS 0x54
//

// Line addresses for LCDs which use
// the Hitachi HD66712U controller chip
/*
#define LCD_PIC_LINE_1_ADDRESS 0x00
#define LCD_PIC_LINE_2_ADDRESS 0x20
#define LCD_PIC_LINE_3_ADDRESS 0x40
#define LCD_PIC_LINE_4_ADDRESS 0x60
*/

int I2C_BYTE;

/* Definition for mjkdz I2C adapter with pot bent over top of chip
// Model tested had I2C Address = 0x20.  Change if yours is different
// Order of pin definitions not important, just be sure all 8 pins are defined and there are no duplicates
// Pin assignment examples: 
//    LCD pin RW is wired directly to I2C chip PCF8574P Output P5 (RW  = I2C_BYTE.5)
//    LCD pin D4 is wired directly to I2C chip PCF8574P Output P0 (D4  = I2C_BYTE.0)
//    and so on...
//
// mjkdz I2C adapter
#define LCD_PIC_addr 0x20
#bit  E   = I2C_BYTE.4
#bit  RW  = I2C_BYTE.5
#bit  RS  = I2C_BYTE.6
#bit  D4  = I2C_BYTE.0
#bit  D5  = I2C_BYTE.1
#bit  D6  = I2C_BYTE.2
#bit  D7  = I2C_BYTE.3
#bit  BL  = I2C_BYTE.7
#define  BL_ON   0     // 1 for POSITIVE control, 0 for NEGATIVE control
*/

/* Notes for Sainsmart LCD_PIC I2C adapter
// Model tested had I2C Address = 0x3F.  Change if yours is different
// Enter P0 bit from IC PCF8574P to LCD connections
// Enter whether Backlight ON is 1 (POSITIVE control) or 0 (NEGATIVE control)
// Note: Regarding the Sainsmart I2C adapter, output-P3 (pin-7) from IC 
//       PCF8574 is wired such that if LOW, it shuts off
//       the LCD's LED backlight.  P3 doesn't seem to affect any other 
//       operations since P3 is not wired to any of the HD44780 
//       data inputs.  Therefore P3 must be HIGH on all write operations. 
//
//    P7 - D7        P3 - BACKLIGHT LED
//    P6 - D6        P2 - E
//    P5 - D5        P1 - R/W
//    P4 - D4        P0 - RS
*/


// Sainsmart I2C adapter LCD2004
#define  LCD_PIC_addr 0x3F
#bit     E   =    I2C_BYTE.2
#bit     RW  =    I2C_BYTE.1
#bit     RS  =    I2C_BYTE.0
#bit     D4  =    I2C_BYTE.4
#bit     D5  =    I2C_BYTE.5
#bit     D6  =    I2C_BYTE.6
#bit     D7  =    I2C_BYTE.7
#bit     BL  =    I2C_BYTE.3
#define  BL_ON    1     // 1 for POSITIVE control, 0 for NEGATIVE control
//

// Following line creates appropriate address byte for I2C interface (shift left 1-bit)
#define LCD_PIC_I2C_ADDR (LCD_PIC_addr<<1)

#separate
void LCD_PIC_I2C_Write_Generic(char text[]); 

#separate
void LCD_PIC_I2C_Init_Generic();

#separate
void LCD_PIC_I2C_GOTOXY_Generic(int8 x, int8 y);

#separate
void LCD_PIC_I2C_Putc_Generic(char c);

#separate
void LCD_PIC_I2C_Send_Byte_Generic(int8 address, int8 n);

#separate
void LCD_PIC_I2C_Write_Upper_Nibble_Generic(int u);

#separate
void LCD_PIC_I2C_Write_Lower_Nibble_Generic(int l);

int lcd_j,lcd_max,lcd_tempbyte,n,u,l;
int lcd_fixed_delay_us=50;  // Fixed delay to allow LCD to process I2C instruction
int LCD_PIC_line;
unsigned char lcd_buffer[20];


//===================================
void LCD_PIC_I2C_Write_Generic(char text[]) 
{
// Writes a string text[] to LCD via I2C
   RS  = 1;
   RW  = 0;
   E   = 0;
   BL  = BL_ON;
   
   i2c_start(LCDI2C);
   i2c_write(LCDI2C,LCD_PIC_I2C_ADDR);

   lcd_max=strlen(text);
   for(lcd_j=0; lcd_j<lcd_max; ++lcd_j)
   {
      lcd_tempbyte=text[lcd_j];
      
// Send upper nibble   
      LCD_PIC_I2C_Write_Upper_Nibble_Generic(lcd_tempbyte);
      
// Send lower nibble
      LCD_PIC_I2C_Write_Lower_Nibble_Generic(lcd_tempbyte);
   }
   i2c_stop(LCDI2C);
}

//===================================
void LCD_PIC_I2C_GOTOXY_Generic(int8 x, int8 y)
{
int8 address;

   switch(y)
     {
      case 1:
        address = LCD_PIC_LINE_1_ADDRESS;
        break;
   
      case 2:
        address = LCD_PIC_LINE_2_ADDRESS;
        break;
   
      case 3:
        address = LCD_PIC_LINE_3_ADDRESS;
        break;
   
      case 4:
        address = LCD_PIC_LINE_4_ADDRESS;
        break;
   
      default:
        address = LCD_PIC_LINE_1_ADDRESS;
        break;     
     }

   address += x-1;
   LCD_PIC_I2C_Send_Byte_Generic(0, 0x80 | address);
}

//===================================
void LCD_PIC_I2C_Putc_Generic(char c)
{
// Writes a char 'c' to LCD via I2C to include some common formats
   switch(c)
   {
      case '\f':
         LCD_PIC_I2C_Send_Byte_Generic(0x00,0x01);
         LCD_PIC_line = 1;
         delay_ms(lcd_fixed_delay_us);
         break;
      
      case '\n':
          LCD_PIC_I2C_Send_Byte_Generic(0x01, ++LCD_PIC_line);
          break;
      
      case '\b':
          LCD_PIC_I2C_Send_Byte_Generic(0x00,0x10);
          break;
      
      default:
          LCD_PIC_I2C_Send_Byte_Generic(0x01,c);
          break;
   }
}

//===================================
void LCD_PIC_I2C_Send_Byte_Generic(int8 address, int8 n)
{
// HD44780U LCD Address Type (RS): Command=0, Data=1
//
   IF(address)
   {
      RS=1;   // Data
   }
   ELSE
   {
      RS=0;   // Command
   }
 
   RW  = 0;
   E   = 0;
   BL  = BL_ON;

   i2c_start(LCDI2C);
   i2c_write(LCDI2C,LCD_PIC_I2C_ADDR);

// Send upper nibble   
   LCD_PIC_I2C_Write_Upper_Nibble_Generic(n);
      
// Send lower nibble
   LCD_PIC_I2C_Write_Lower_Nibble_Generic(n);
   
   i2c_stop(LCDI2C);
//   delay_ms(lcd_fixed_delay_us);
}

//===================================
void LCD_PIC_I2C_Init_Generic()
{  
// Initialization commands for Hitachi HD44780U LCD Display
//
   delay_ms(300);                               // Let LCD power up

   i2c_start(LCDI2C);                           // Begin transfer, claim I2C BUS
   i2c_write(LCDI2C,LCD_PIC_I2C_ADDR);          // Tell all I2C devices you are talking to LCD_PIC_I2C_ADDR

// Following bytes are all Command bytes, i.e. address = 0x00
   LCD_PIC_I2C_Send_Byte_Generic(0x00, 0x03);   // Write Nibble 0x03 three times (per HD44780U initialization spec)
   delay_us(lcd_fixed_delay_us);                // (per HD44780U initialization spec)
   LCD_PIC_I2C_Send_Byte_Generic(0x00, 0x03);   //
   delay_us(lcd_fixed_delay_us);                // (per HD44780U initialization spec)
   LCD_PIC_I2C_Send_Byte_Generic(0x00, 0x03);   //
   delay_us(lcd_fixed_delay_us);                // (per HD44780U initialization spec)
   LCD_PIC_I2C_Send_Byte_Generic(0x00, 0x02);   // Write Nibble 0x02 once (per HD44780U initialization spec)
   delay_us(lcd_fixed_delay_us);                // (per HD44780U initialization spec)
   LCD_PIC_I2C_Send_Byte_Generic(0x00, 0x28);   // Set mode: 4-bit, 2+lines, 5x8 dots
   delay_us(lcd_fixed_delay_us);                // (per HD44780U initialization spec) 
   LCD_PIC_I2C_Send_Byte_Generic(0x00, 0x0C);   // Display ON
   delay_us(lcd_fixed_delay_us);                // (per HD44780U initialization spec) 
   LCD_PIC_I2C_Send_Byte_Generic(0x00, 0x01);   // Clear display
   delay_ms(lcd_fixed_delay_us);                // (per HD44780U initialization spec) 
   LCD_PIC_I2C_Send_Byte_Generic(0x00, 0x06);   // Set cursor to increment
   delay_us(lcd_fixed_delay_us);                // (per HD44780U initialization spec) 
   i2c_stop(LCDI2C);                            // Terminate I2C transfer
}

//===================================
void LCD_PIC_I2C_Write_Upper_Nibble_Generic(int u) 
{
// Send upper nibble
   IF(bit_test(u,7))
      D7=1;
   ELSE
      D7=0;
      
   IF(bit_test(u,6))
      D6=1;
   ELSE
      D6=0;

   IF(bit_test(u,5))
      D5=1;
   ELSE
      D5=0;
      
   IF(bit_test(u,4))
      D4=1;
   ELSE
      D4=0;
             
   E = 0;    
   i2c_write(LCDI2C,I2C_BYTE);
   E = 1;    
   i2c_write(LCDI2C,I2C_BYTE);
   E = 0;    
   i2c_write(LCDI2C,I2C_BYTE);
}

//===================================
void LCD_PIC_I2C_Write_Lower_Nibble_Generic(int l) 
{
// Send lower nibble
   IF(bit_test(l,3))
      D7=1;
   ELSE
      D7=0;
      
   IF(bit_test(l,2))
      D6=1;
   ELSE
      D6=0;

   IF(bit_test(l,1))
      D5=1;
   ELSE
      D5=0;
      
   IF(bit_test(l,0))
      D4=1;
   ELSE
      D4=0;

   E = 0;    
   i2c_write(LCDI2C,I2C_BYTE);
   E = 1;    
   i2c_write(LCDI2C,I2C_BYTE);
   E = 0;    
   i2c_write(LCDI2C,I2C_BYTE);  
}
